// CVS ID : $Id: ECCV.cc,v 1.12 2010/12/05 20:14:48 wuthrich Exp $
// ---------------------------------------------------------------
//
// Definition of the class ECCV
//
// (c) Rolf Wuthrich
//     2010 Concordia University
//
// author:  Rolf Wuthrich
// email:   wuthrich@encs.concordia.ca
// version: v1 ($Revision: 1.12 $)
//
// This software is copyright under the BSD license
//  
// ---------------------------------------------------------------
// history:
// ---------------------------------------------------------------
// Date       Modification                     Author
// ---------------------------------------------------------------
// 31.03.2010 Creation                         R. Wuthrich
// ---------------------------------------------------------------

#include <cmath>
#include "ECCV.h"

using namespace std;


// -------------------------------------------------------------------------
//
// Tcl interface


// Command to create a new CV object
//

int ECTk_CvCmd(ClientData clientData, Tcl_Interp *interp, 
	       int objc,Tcl_Obj *const objv[])
{
  ECCV* cv = new ECCV(interp);
  return ECTk_CreateECProcessCmd(interp,clientData,cv,objc,objv);
}


// -------------------------------------------------------------------------


// -----------------------------------------------------------------------
//
// class ECCV : cyclic voltammetry
//


// ---- HANDLEEVENT : define how ECCV reacts to external events
//

void ECCV::handleEvent(ECEvent* ev)
{
  ECProcedure::handleEvent(ev);

  // if not Active nothing to do
  if (status!=Active) return;

  // checks for a scan rate of 0
  if (scanRate==0)
    {
      Tcl_Eval(interp,"cons \"\t-- ERROR can't run with a scan rate of 0\"");
      terminate();
      ev->setCmdEvent(CmdProcessTerminated,this);
      return;
    }

  char str[80];
  static int scan_n=1;

  // checks if the procedure was activated
  if (ev->cmd()==CmdProcedureActivated) scan_n=1;

  // checks if all scans where performed
  if (timeSinceActivated()>=2*N*abs(E2-E1)/scanRate)
    {
      terminate();
      ev->setCmdEvent(CmdProcessTerminated,this);
    }

  // checks if a new scan starts
  if (timeSinceActivated()>=2*(scan_n-1)*abs(E2-E1)/scanRate)
    {
      sprintf(str,"c%d",scan_n);
      setMarkStr(str);
      if (verbose==1)
	{
	  sprintf(str,"cons \"\t -- running scan # %d\"",scan_n); 
	  Tcl_Eval(interp,str);
	}
      scan_n++;
    }
}


// ---- POTENTIAL : Computes E(t)
//                  Implements triangular signals

int sign(double d)
{
  if (d>0)
    return 1;
  else 
    return -1;
}

double ECCV::potential()
{
  double T=abs(E2-E1)/scanRate;
  double t=timeSinceActivated();
  int n=floor(t/T);
  t=t-2*floor(t/T/2)*T;
  if (n % 2 == 0)
    return E1+sign(E2-E1)*scanRate*t;
  else
    return E2-sign(E2-E1)*scanRate*(t-T);
}


// ---- INFO : returns a Tcl_DString containing relevant info about 
//             the electrochemical procedure

Tcl_DString* ECCV::info(const char* frontStr)
{
  ECProcedure::info(frontStr);
  addInfo(frontStr,"scan rate (V/s) = ",scanRate);
  addInfo(frontStr,"initial potential (V) = ",E1);
  addInfo(frontStr,"final potential (V) = ",E2);
  addInfo(frontStr,"number of cycles = ",N);
  return infoStr;
}
