// CVS ID : $Id: ECData.cc,v 1.9 2011/03/08 01:33:28 wuthrich Exp $
// ---------------------------------------------------------------
//
// Definition of the class ECData
//
// (c) Rolf Wuthrich
//     2010 Concordia University
//
// author:  Rolf Wuthrich
// email:   wuthrich@encs.concordia.ca
// version: v1 ($Revision: 1.9 $)
//
// This software is copyright under the BSD license
//  
// ---------------------------------------------------------------
// history:
// ---------------------------------------------------------------
// Date       Modification                     Author
// ---------------------------------------------------------------
// 12.07.2010 Creation                         R. Wuthrich
// ---------------------------------------------------------------

#include <tcl.h>
#include <fstream>
#include <cmath>
#include <iostream>
#include <cstring>
#include <algorithm>
#include "ECData.h"

using namespace std;


// -----------------------------------------------------------------------
//
// class ECRecord : a class for holding electrochemical data
//


// ---- CONSTRUCTORS
//

ECRecord::ECRecord(double T, double potential, double current, 
		   double charge, double PosCharge, double NegCharge,
		   const char* Mark):
  nextECRecord(NULL),
  t(T), E(potential), I(current), Q(charge), Qpos(PosCharge), Qneg(NegCharge)
{
  prnStr = new Tcl_DString;
  Tcl_DStringInit(prnStr);
  mark = new Tcl_DString;
  Tcl_DStringInit(mark);
  Tcl_DStringAppend(mark,Mark,-1);
}

ECRecord::ECRecord(ECRecord* ecrec)
{
  t=ecrec->t;
  E=ecrec->E;
  I=ecrec->I;
  Q=ecrec->Q;
  Qpos=ecrec->Qpos;
  Qneg=ecrec->Qneg;
  nextECRecord=NULL;
  prnStr = new Tcl_DString;
  Tcl_DStringInit(prnStr);
  mark = new Tcl_DString;
  Tcl_DStringInit(mark);
  Tcl_DStringAppend(mark,Tcl_DStringValue(ecrec->mark),-1);
}


// --- DESTRUCTOR
//

ECRecord::~ECRecord()
{
  Tcl_DStringFree(prnStr);
  delete prnStr;
  Tcl_DStringFree(mark);
  delete mark;
}


// --- print : returns a Tcl_DString with the data in the record
//

Tcl_DString* ECRecord::print(const char* separator)
{
  Tcl_Interp *interp;
  interp=Tcl_CreateInterp();  

  Tcl_DStringFree(prnStr);

  char str[15];

  sprintf(str,"%9.3g ",t);
  Tcl_DStringAppend(prnStr,str,-1);
  Tcl_DStringAppend(prnStr,separator,-1);

  sprintf(str,"%9.3g ",E);
  Tcl_DStringAppend(prnStr,str,-1);
  Tcl_DStringAppend(prnStr,separator,-1);

  sprintf(str,"%9.3g ",I);
  Tcl_DStringAppend(prnStr,str,-1);
  Tcl_DStringAppend(prnStr,separator,-1);

  sprintf(str,"%9.3g ",Q);
  Tcl_DStringAppend(prnStr,str,-1);
  Tcl_DStringAppend(prnStr,separator,-1);

  sprintf(str,"%9.3g ",Qpos);
  Tcl_DStringAppend(prnStr,str,-1);
  Tcl_DStringAppend(prnStr,separator,-1);

  sprintf(str,"%9.3g ",Qneg);
  Tcl_DStringAppend(prnStr,str,-1);
  Tcl_DStringAppend(prnStr,separator,-1);

  sprintf(str," %-8.8s",Tcl_DStringValue(mark));
  Tcl_DStringAppend(prnStr,str,-1);

  Tcl_DeleteInterp(interp);

  return prnStr;
}


// --- cpyECRecToTcl : copies the ECRec to the associated Tcl namespace
//

void ECRecord::cpyECRecToTcl(Tcl_Interp* interp, Tcl_DString* NameSpace)
{
  int len = Tcl_DStringLength(NameSpace);

  Tcl_SetVar(interp,Tcl_DStringAppend(NameSpace,"::mark",-1),
	     Tcl_DStringValue(mark),TCL_NAMESPACE_ONLY);

  Tcl_DStringSetLength(NameSpace,len);
  Tcl_UnlinkVar(interp,Tcl_DStringAppend(NameSpace,"::t",-1));
  Tcl_LinkVar(interp,Tcl_DStringValue(NameSpace),(char *)&t,TCL_LINK_DOUBLE);
  Tcl_DStringSetLength(NameSpace,len);
  Tcl_UnlinkVar(interp,Tcl_DStringAppend(NameSpace,"::E",-1));
  Tcl_LinkVar(interp,Tcl_DStringValue(NameSpace),(char *)&E,TCL_LINK_DOUBLE);
  Tcl_DStringSetLength(NameSpace,len);
  Tcl_UnlinkVar(interp,Tcl_DStringAppend(NameSpace,"::I",-1));
  Tcl_LinkVar(interp,Tcl_DStringValue(NameSpace),(char *)&I,TCL_LINK_DOUBLE);
  Tcl_DStringSetLength(NameSpace,len);
  Tcl_UnlinkVar(interp,Tcl_DStringAppend(NameSpace,"::Q",-1));
  Tcl_LinkVar(interp,Tcl_DStringValue(NameSpace),(char *)&Q,TCL_LINK_DOUBLE);
  Tcl_DStringSetLength(NameSpace,len);
  Tcl_UnlinkVar(interp,Tcl_DStringAppend(NameSpace,"::Qpos",-1));
  Tcl_LinkVar(interp,Tcl_DStringValue(NameSpace),(char *)&Qpos,TCL_LINK_DOUBLE);
  Tcl_DStringSetLength(NameSpace,len);
  Tcl_UnlinkVar(interp,Tcl_DStringAppend(NameSpace,"::Qneg",-1));
  Tcl_LinkVar(interp,Tcl_DStringValue(NameSpace),(char *)&Qneg,TCL_LINK_DOUBLE);
  Tcl_DStringSetLength(NameSpace,len);
}



// -----------------------------------------------------------------------
//
// class ECData : abstract class able to hold EC data (a list of ECRecord)
//


// ---- CONSTRUCTOR
//

ECData::ECData(const char* Name, Tcl_Interp *Interp) : 
  ECObj(Name,"",Interp),
  n(0),dt_out(0.1),
  firstECRecord(NULL),lastECRecord(NULL),currentECRecord(NULL)
{
  markStr=Tcl_Alloc(1);
  strcpy(markStr,"");
  cut       = new Tcl_DString;
  xCoord    = new Tcl_DString;
  yCoord    = new Tcl_DString;
  plotStyle = new Tcl_DString;
  Tcl_DStringInit(cut);
  Tcl_DStringAppend(cut,"return 1",-1);
  Tcl_DStringInit(xCoord);
  Tcl_DStringAppend(xCoord,"$t",-1);
  Tcl_DStringInit(yCoord);
  Tcl_DStringAppend(yCoord,"$E",-1);
  Tcl_DStringInit(plotStyle);
  Tcl_DStringAppend(plotStyle,"with lines title '",-1);
  Tcl_DStringAppend(plotStyle,Tcl_DStringValue(TclCmd),-1);
  Tcl_DStringAppend(plotStyle,"'",-1);

  addOption("-stat","returns statistics about records");
  addOption("-readFirstRecord","returns first record");
  addOption("-readNextRecord","returns next record");
  addOption("-cut","sets cut condition");
  addOption("-dtOut",&dt_out,"returns/sets time interval between two outputs");
  addOption("-loop","loops over all records and executes a user script");
  addOption("-print","prints records to stdout");
  addOption("-save","saves records to a file");
  addOption("-plot","plots records");
  addOption("-x",xCoord,"defines X coordinate for ploting");
  addOption("-y",yCoord,"defines Y coordinate for ploting");
  addOption("-style",plotStyle,"defines plot style");
  addOption("-max","returns record with maximal current");
  addOption("-min","returns record with minimal current");
  addOption("-mean","returns avarage values in the Tcl array mean");
  addOption("-meanI","returns avarage current");
  addOption("-meanE","returns avarage potential");
  addOption("-cvQ","returns charge for a cv experiment in Tcl array cvQ");

  nsPtr=Tcl_CreateNamespace(interp,Tcl_DStringValue(TclCmd), NULL, NULL);
}


// ---- DESTRUCTOR
//

ECData::~ECData()
{
  removeRecords();
  Tcl_Free(markStr);
  Tcl_DStringFree(cut);
  Tcl_DStringFree(xCoord);
  Tcl_DStringFree(yCoord);
  Tcl_DStringFree(plotStyle);
  delete cut;
  delete xCoord;
  delete yCoord;
  delete plotStyle;
  Tcl_DeleteNamespace(nsPtr);
}


// ---- SETCUT : parse the the "-cut" option
//

int ECData::setCut(int objc,Tcl_Obj *const objv[])
{
  int i=1;
  int index=-1;
  Tcl_DStringInit(cut);
  Tcl_DStringAppend(cut,"return 1",-1);

  while (i<objc)
    {
      Tcl_GetIndexFromObj(NULL,objv[i],optsTbl(),"",0,&index);
      if (index==getOptionIndex("-cut"))
	{
	  if (i+1>=objc) 
	    {
	      Tcl_WrongNumArgs(interp,i+1,objv,"cutCondition");
	      return TCL_ERROR;
	    }
	  Tcl_DStringInit(cut);
	  Tcl_DStringAppend(cut,"namespace eval ",-1);
	  Tcl_DStringAppend(cut,Tcl_DStringValue(TclCmd),-1);
	  Tcl_DStringAppend(cut," {if {",-1);
	  Tcl_DStringAppend(cut,Tcl_GetString(objv[i+1]),-1);
	  Tcl_DStringAppend(cut,"} {return 1} else {return -1}}",-1);
	  return TCL_OK;
	}
      i++;
    }
  return TCL_OK;
}


// ---- SETXY : parse the the "-x" and "-y" option
//

int ECData::setXY(int objc,Tcl_Obj *const objv[])
{
  int i=1;
  int index=-1;

  while (i<objc)
    {
      Tcl_GetIndexFromObj(NULL,objv[i],optsTbl(),"",0,&index);
      if (index==getOptionIndex("-x"))
	{
	  if (i+1>=objc) 
	    {
	      Tcl_WrongNumArgs(interp,i+1,objv,"xCoordinate");
	      return TCL_ERROR;
	    }
	  Tcl_DStringFree(xCoord);
	  Tcl_DStringAppend(xCoord,Tcl_GetString(objv[i+1]),-1);
	  i++;
	}

      if (index==getOptionIndex("-y"))
	{
	  if (i+1>=objc) 
	    {
	      Tcl_WrongNumArgs(interp,i+1,objv,"yCoordinate");
	      return TCL_ERROR;
	    }
	  Tcl_DStringFree(yCoord);
	  Tcl_DStringAppend(yCoord,Tcl_GetString(objv[i+1]),-1);
	  i++;
	}

      if (index==getOptionIndex("-style"))
	{
	  if (i+1>=objc) 
	    {
	      Tcl_WrongNumArgs(interp,i+1,objv,"plotStyle");
	      return TCL_ERROR;
	    }
	  Tcl_DStringFree(plotStyle);
	  Tcl_DStringAppend(plotStyle,Tcl_GetString(objv[i+1]),-1);
	  i++;
	}
      i++;
    }
  return TCL_OK;
}


// ---- CMD : process a potential command option (# i of objv[])
//            searches the options table 
//            if a match is found processes the command option and
//            returns the index from the options table
//            if no match returns -1

// ---- auxiliary function for cmd

void ECData::parseMarks(int &i,int objc,Tcl_Obj* const objv[],
			char* &begin, char* &end)
{
  if ((i+1<=objc-1)&&(Tcl_StringMatch(Tcl_GetString(objv[i+1]),"-*")==0))
    {
      begin=Tcl_GetString(objv[i+1]);
      if ((i+2<=objc-1)&&(Tcl_StringMatch(Tcl_GetString(objv[i+2]),"-*")==0))
	{
	  end=Tcl_GetString(objv[i+2]);
	  i++;
	}
      else
	end=Tcl_GetString(objv[i+1]);
      i++;
    }
}


int ECData::cmd(int &i,int objc,Tcl_Obj *const objv[])
{
  // first executes the command options defined in ECProcess
  int index=ECObj::cmd(i,objc,objv);

  char empty[]="";
  char* begin=empty;
  char* end=empty;

  if (index==getOptionIndex("-cut"))
    {
      if (i+1>=objc) 
	{
	  Tcl_WrongNumArgs(interp,i+1,objv,"cutCondition");
	  return -1;
	}
      i=i+2;
    }

  if (index==getOptionIndex("-print"))
  {
    if (setCut(objc,objv)!=TCL_OK) return -1;
    parseMarks(i,objc,objv,begin,end);
    if (scan(begin,end)!=TCL_OK)
      {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp,"invalid startMark",NULL);
	return -1;
      }
    i++;
  }

  if (index==getOptionIndex("-loop"))
    {
      if (i+1>=objc) 
	{
	  Tcl_WrongNumArgs(interp,i+1,objv,"loopScript ?startMark? ?endMark?");
	  return -1;
	}
      char* loopScript=Tcl_GetString(objv[i+1]);
      if (setCut(objc,objv)!=TCL_OK) return -1;
      i++;
      parseMarks(i,objc,objv,begin,end);
      loop(loopScript,begin,end);
      i++;
    }

  if (index==getOptionIndex("-readFirstRecord"))
    {	
      currentECRecord=firstECRecord;
      if (currentECRecord!=NULL)
	currentECRecord->cpyECRecToTcl(interp,TclCmd); 
      i++;
    }

  if (index==getOptionIndex("-readNextRecord"))
    {
      if (currentECRecord!=NULL)
	{
	  currentECRecord=currentECRecord->getNextECRecord();
	  if (currentECRecord!=NULL)
	    currentECRecord->cpyECRecToTcl(interp,TclCmd); 
	  else
	    Tcl_AppendResult(interp,"-1",NULL);
	}
      i++;
    }

  if (index==getOptionIndex("-stat"))
    {
      recordsStat();
      i++;
    }

  if (index==getOptionIndex("-save"))
    {
      if (i+1>objc-1)
	{
	  Tcl_WrongNumArgs(interp,i+1,objv,"filename ?startMark? ?endMark?");
	  return -1;
	}
      if (setCut(objc,objv)!=TCL_OK) return -1;
      char* fileName=Tcl_GetString(objv[i+1]);
      i++;
      parseMarks(i,objc,objv,begin,end);
      saveRecords(fileName,ios::trunc,begin,end);
      Tcl_AppendResult(interp,"saved ",fileName,NULL);
      i++;
    }

  if (index==getOptionIndex("-plot"))
    {
      parseMarks(i,objc,objv,begin,end);
      if (setCut(objc,objv)!=TCL_OK) return -1;
      if (setXY(objc,objv)!=TCL_OK) return -1;
      Tcl_Eval(interp,"set plotFile [open .ectmpfile w]");
      Tcl_DString* loopScript = new Tcl_DString;
      Tcl_DStringInit(loopScript);
      Tcl_DStringAppend(loopScript,"puts $plotFile \"[expr {",-1);
      Tcl_DStringAppend(loopScript,Tcl_DStringValue(xCoord),-1);
      Tcl_DStringAppend(loopScript,"}] [expr {",-1);
      Tcl_DStringAppend(loopScript,Tcl_DStringValue(yCoord),-1);
      Tcl_DStringAppend(loopScript,"}]\"",-1);
      loop(Tcl_DStringValue(loopScript),begin,end);
      Tcl_DStringFree(loopScript);
      Tcl_DStringAppend(loopScript,"plotGnuplot .ectmpfile ",-1);
      Tcl_DStringAppend(loopScript,Tcl_DStringValue(plotStyle),-1);
      Tcl_Eval(interp,Tcl_DStringValue(loopScript));
      Tcl_Eval(interp,"close $plotFile");
      Tcl_DStringFree(loopScript);
      delete loopScript;
      i++;
    }

  if (index==getOptionIndex("-max"))
    {
      parseMarks(i,objc,objv,begin,end);
      if (setCut(objc,objv)!=TCL_OK) return -1;
      if (findMax(begin,end)!=TCL_OK)
	{
	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp,"invalid startMark",NULL);
	  return -1;
	}
      i++;
    }

  if (index==getOptionIndex("-min"))
    {
      parseMarks(i,objc,objv,begin,end);
      if (setCut(objc,objv)!=TCL_OK) return -1;
      if (findMin(begin,end)!=TCL_OK)
	{
	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp,"invalid startMark",NULL);
	  return -1;
	}
      i++;
    }

  if (index==getOptionIndex("-mean"))
    {
      parseMarks(i,objc,objv,begin,end);
      if (setCut(objc,objv)!=TCL_OK) return -1;
      if (mean(begin,end)!=TCL_OK)
	{
	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp,"invalid startMark",NULL);
	  return -1;
	}
      i++;
    }

  if (index==getOptionIndex("-meanI"))
    {
      parseMarks(i,objc,objv,begin,end);
      if (setCut(objc,objv)!=TCL_OK) return -1;
      if (meanI(begin,end)!=TCL_OK)
	{
	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp,"invalid startMark",NULL);
	  return -1;
	}
      i++;
    }

  if (index==getOptionIndex("-meanE"))
    {
      parseMarks(i,objc,objv,begin,end);
      if (setCut(objc,objv)!=TCL_OK) return -1;
      if (meanE(begin,end)!=TCL_OK)
	{
	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp,"invalid startMark",NULL);
	  return -1;
	}
      i++;
    }

  if (index==getOptionIndex("-cvQ"))
    {
      if (i+1>objc-1)
	{
	  Tcl_WrongNumArgs(interp,i+1,objv,"scanRate ?startMark? ?endMark?");
	  return -1;
	}
      double scanRate;
      if (Tcl_GetDoubleFromObj(interp,objv[i+1],&scanRate)!=TCL_OK) return -1;
      i++;
      parseMarks(i,objc,objv,begin,end);
      if (setCut(objc,objv)!=TCL_OK) return -1;
      if (cvQ(scanRate,begin,end)!=TCL_OK)
	{
	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp,"invalid startMark",NULL);
	  return -1;
	}
      i++;
    }

  return index;
}

 
// ---- INFO : returns a Tcl_DString containing relevant info about content
//             Has to be defined by child

Tcl_DString* ECData::info(const char* frontStr)
{
  ECObj::info(frontStr);
}


// ---- GETECRECORD : returns the first record with mark=Mark 
//

ECRecord*   ECData::getECRecord(const char* Mark)
{
  ECRecord* rec=firstECRecord;
  if (rec==NULL) return NULL;

  // locate the record
  if (strcmp(Mark,"")!=0)
    while ((rec!=NULL)&&(strcmp(Mark,Tcl_DStringValue(rec->getMark()))!=0))
      rec=rec->getNextECRecord();
  return rec;
}


// ---- ADDECRECORD : adds a new ECRecord into the list
//

void ECData::addECRecord(double T, double potential, double current, 
           double charge, double PosCharge, double NegCharge)
{
  ECRecord* rec = 
    new ECRecord(T,potential,current,charge,PosCharge,NegCharge,markStr);
  n++;

  if (lastECRecord==NULL)
    {
      lastECRecord=rec;
      firstECRecord=rec;
      currentECRecord=rec;
    }
  else
    {
      lastECRecord->setNextECRecord(rec);
      lastECRecord=rec;
    }
}


// ---- REMOVERECORDS : removes all ECRecord in the list
//

void ECData::removeRecords()
{
  ECRecord* rec=firstECRecord;

  while(rec!=NULL)
    {
      rec=rec->getNextECRecord();
      delete firstECRecord;
      firstECRecord=rec;
    }

  firstECRecord=NULL;
  lastECRecord=NULL;
  currentECRecord=NULL;

  n=0;
}


// ---- RECORDSSTAT : produces stats about the records in the list
//

void ECData::recordsStat()
{
  ECRecord* rec=firstECRecord;
  double t1=0.0;
  double t2;
  double diff;
  double sum     = 0.0;
  double sum_sqr = 0.0;
  double max     = 0.0;
  double min     = 0.0;
  ofstream file;
  file.open("histo.txt",ios::trunc);

  while(rec!=NULL)
    {
      t2=rec->getT();
      diff=(t2-t1)*1000;
      file << diff << "\n";
      if (diff>max) max=diff;
      if (diff<min) min-diff;
      sum=sum+diff;
      sum_sqr=sum_sqr+diff*diff;
      t1=t2;
      rec=rec->getNextECRecord();
    }

  file.close();
  cout << "# records = " << n << "\n";
  cout << "<dt> = " << sum/n << " ms\n";
  cout << "std dev = " << sqrt((sum_sqr-sum*sum/n))/(n-1) << " ms\n";
  cout << "min = " << min << " ms\n";
  cout << "max = " << max << " ms\n";
}


// ---- GETNBRECRECORDS : returns n
//                        can be used to distinguish from an ECProcess

int ECData::getNbrECRecords()
{
  return n;
}


// ---- LOOP : loops over records between startMark and endMark
//
//             if startMark = "" starts from the first record
//

int ECData::loop(const char* loopScript,
		 const char* startMark, const char* endMark)
{
  ECRecord* rec=getECRecord(startMark);
  if (strcmp(endMark,"")==0) endMark=startMark;

  int last=0;

  while (rec!=NULL)
    {
      if ((last==1)&&(strcmp(endMark,Tcl_DStringValue(rec->getMark()))!=0))
	break;
      rec->cpyECRecToTcl(interp,TclCmd);
      Tcl_Eval(interp,Tcl_DStringValue(cut));
      if (strcmp(Tcl_GetStringResult(interp),"1")==0)
	{ 
	  Tcl_VarEval(interp,"namespace eval ",
		      Tcl_DStringValue(TclCmd),
		      " {",
		      loopScript,"}",NULL);
	}
      if (strcmp(endMark,Tcl_DStringValue(rec->getMark()))==0) last=1;
      rec=rec->getNextECRecord();
    }
}


// ---- SAVERECORDS : save all ECRecords in the list to a file
//
//                    if startMark = "" starts from the first record
//

int ECData::saveRecords(const char* fileName, ios_base::openmode mode,
			const char* startMark, const char* endMark)
{
  ofstream file;
  file.open(fileName,mode);
  char str[8];
  strcpy(str,Tcl_GetVar(interp,"::ectk::commentStr",0));
  file << Tcl_DStringValue(info(str)) << "\n";
  file << str << "---------------------------------------------------\n";
  file << str << "User : " << Tcl_GetVar(interp,"::ectk::user",0) << "\n";
  time_t time_now;
  time(&time_now);
  file << str << "Date : " << ctime(&time_now);
  file << str << "---------------------------------------------------\n";
  file << str << "t [s]\tE [V]\tI [A]\tQ [C]\tQ+ [C]\tQ- [C]\tMark\n";
  file << str << "---------------------------------------------------\n";

  ECRecord* rec=getECRecord(startMark);
  if (strcmp(endMark,"")==0) endMark=startMark;

  double next_out=0.0;
  int last=0;

  while (rec!=NULL)
    {
      if ((last==1)&&(strcmp(endMark,Tcl_DStringValue(rec->getMark()))!=0))
	break;
      if (rec->getT()>=next_out)
	{
	  rec->cpyECRecToTcl(interp,TclCmd);
	  Tcl_Eval(interp,Tcl_DStringValue(cut));
	  if (strcmp(Tcl_GetStringResult(interp),"1")==0)
	    { 
	      file << Tcl_DStringValue(rec->print()) << "\n";
	      next_out=next_out+dt_out;
	    }
	}
      if (strcmp(endMark,Tcl_DStringValue(rec->getMark()))==0) last=1;
      rec=rec->getNextECRecord();
    }
  file.close();
  return TCL_OK;
}


// ---- SCAN : sends the ECRecords to stdoutput
//

int ECData::scan(const char* startMark, const char* endMark)
{
  static const char* sep = "+----------+----------+----------+----------+----------+----------+----------+\n";

  ECRecord* rec=getECRecord(startMark);

  if (rec==NULL) return TCL_ERROR;
  if (strcmp(endMark,"")==0) endMark=startMark;

  cout << sep;
  cout << "| t [s]    | E [V]    | I [A]    | Q [C]    | Q+ [C]   | Q- [C]   | Mark     |\n";
  cout << sep;

  int j=1;
  int last = 0;
  Tcl_DString* str;
  str = new Tcl_DString;
  Tcl_DStringInit(str);
  Tcl_Channel inChan = Tcl_GetStdChannel(TCL_STDIN);

  while (rec!=NULL)
    {
      if ((last==1)&&(strcmp(endMark,Tcl_DStringValue(rec->getMark()))!=0))
	break;

      rec->cpyECRecToTcl(interp,TclCmd);
      Tcl_Eval(interp,Tcl_DStringValue(cut));
      if (strcmp(Tcl_GetStringResult(interp),"1")!=0) 
	{
	  rec=rec->getNextECRecord();
	  continue;
	}
      cout << "|" << Tcl_DStringValue(rec->print("|")) << " |\n";

      j++;
      if (j==16)
	{
	  cout << sep;
	  cout << "more ? (y/n) ";
	  cout.flush();
	  Tcl_DStringInit(str);
	  while (Tcl_Gets(inChan,str)<=0)
	    Tcl_DoOneEvent(TCL_DONT_WAIT);
	  if (strcmp((Tcl_DStringValue(str)),"n")==0) break; else j=1;
	}

      if (strcmp(endMark,Tcl_DStringValue(rec->getMark()))==0) last=1;
      rec=rec->getNextECRecord();

    }

  if (j<16) cout << sep;
  Tcl_DStringFree(str);
  delete str;
  return TCL_OK;
}


// ---- SETMARKSTR : sets markStr
//

void ECData::setMarkStr(const char* Mark)
{
  Tcl_Realloc(markStr,strlen(Mark)+1);
  strcpy(markStr,Mark);
}


// ---- dtOut : returns dt_out
//

double ECData::dtOut()
{
  return dt_out;
}


// ---- findMax : finds max I
//

int ECData::findMax(const char* startMark, const char* endMark)
{
  ECRecord* rec=getECRecord(startMark);
  if (rec==NULL) return TCL_ERROR;

  if (strcmp(endMark,"")==0) endMark=startMark;

  int last = 0;
  double max = rec->getI();
  ECRecord* max_r=rec;

  while (rec!=NULL)
    {
      if ((last==1)&&(strcmp(endMark,Tcl_DStringValue(rec->getMark()))!=0))
	break;

      if (rec->getI()>max)
	{
	  rec->cpyECRecToTcl(interp,TclCmd);
	  Tcl_Eval(interp,Tcl_DStringValue(cut));
	  if (strcmp(interp->result,"1")==0)
	    {
	      max=rec->getI();
	      max_r=rec;
	    }
	}

      if (strcmp(endMark,Tcl_DStringValue(rec->getMark()))==0) last=1;
      rec=rec->getNextECRecord();
    }

  max_r->cpyECRecToTcl(interp,TclCmd);
  char str[TCL_DOUBLE_SPACE];
  Tcl_PrintDouble(interp,max_r->getE(),str);
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,str,NULL);
  return TCL_OK;
}


// ---- findMin : finds min I
//

int ECData::findMin(const char* startMark, const char* endMark)
{
  ECRecord* rec=getECRecord(startMark);
  if (rec==NULL) return TCL_ERROR;

  if (strcmp(endMark,"")==0) endMark=startMark;

  int last = 0;
  double min = rec->getI();
  ECRecord* min_r=rec;

  while (rec!=NULL)
    {
      if ((last==1)&&(strcmp(endMark,Tcl_DStringValue(rec->getMark()))!=0))
	break;

      if (rec->getI()<min)
	{
	  rec->cpyECRecToTcl(interp,TclCmd);
	  Tcl_Eval(interp,Tcl_DStringValue(cut));
	  if (strcmp(interp->result,"1")==0)
	    {
	      min=rec->getI();
	      min_r=rec;
	    }
	}

      if (strcmp(endMark,Tcl_DStringValue(rec->getMark()))==0) last=1;
      rec=rec->getNextECRecord();
    }

  min_r->cpyECRecToTcl(interp,TclCmd);
  char str[TCL_DOUBLE_SPACE];
  Tcl_PrintDouble(interp,min_r->getE(),str);
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,str,NULL);
  return TCL_OK;
}


// ---- mean : computes mean values
//

int ECData::mean(const char* startMark, const char* endMark)
{
  ECRecord* rec=getECRecord(startMark);
  if (rec==NULL) return TCL_ERROR;

  if (strcmp(endMark,"")==0) endMark=startMark;

  int last = 0;
  int n = 0;
  double meanI = 0;
  double meanE = 0;
  double meanQ = 0;
  double meanQpos = 0;
  double meanQneg = 0;

  while (rec!=NULL)
    {
      if ((last==1)&&(strcmp(endMark,Tcl_DStringValue(rec->getMark()))!=0))
	break;

      rec->cpyECRecToTcl(interp,TclCmd);
      Tcl_Eval(interp,Tcl_DStringValue(cut));
      if (strcmp(interp->result,"1")==0)
	{
	  meanI = meanI+rec->getI();
	  meanE = meanE+rec->getE();
	  meanQ = meanQ+rec->getQ();
	  meanQpos = meanQpos+rec->getQpos();
	  meanQneg = meanQneg+rec->getQneg();
	  n++;
	}

      if (strcmp(endMark,Tcl_DStringValue(rec->getMark()))==0) last=1;
      rec=rec->getNextECRecord();
    }

  char str[TCL_DOUBLE_SPACE];
  Tcl_PrintDouble(interp,meanI/n,str);
  Tcl_SetVar2(interp,"mean","I",str,TCL_GLOBAL_ONLY);
  Tcl_PrintDouble(interp,meanE/n,str);
  Tcl_SetVar2(interp,"mean","E",str,TCL_GLOBAL_ONLY);
  Tcl_PrintDouble(interp,meanQ/n,str);
  Tcl_SetVar2(interp,"mean","Q",str,TCL_GLOBAL_ONLY);
  Tcl_PrintDouble(interp,meanQpos/n,str);
  Tcl_SetVar2(interp,"mean","Qpos",str,TCL_GLOBAL_ONLY);
  Tcl_PrintDouble(interp,meanQneg/n,str);
  Tcl_SetVar2(interp,"mean","Qneg",str,TCL_GLOBAL_ONLY);
  Tcl_ResetResult(interp);
  return TCL_OK;
}


// ---- meanI : computes mean I
//

int ECData::meanI(const char* startMark, const char* endMark)
{
  ECRecord* rec=getECRecord(startMark);
  if (rec==NULL) return TCL_ERROR;

  if (strcmp(endMark,"")==0) endMark=startMark;

  int last = 0;
  int n = 0;
  double mean = 0;

  while (rec!=NULL)
    {
      if ((last==1)&&(strcmp(endMark,Tcl_DStringValue(rec->getMark()))!=0))
	break;

      rec->cpyECRecToTcl(interp,TclCmd);
      Tcl_Eval(interp,Tcl_DStringValue(cut));
      if (strcmp(interp->result,"1")==0)
	{
	  mean = mean+rec->getI();
	  n++;
	}

      if (strcmp(endMark,Tcl_DStringValue(rec->getMark()))==0) last=1;
      rec=rec->getNextECRecord();
    }

  mean = mean/n;
  char str[TCL_DOUBLE_SPACE];
  Tcl_PrintDouble(interp,mean,str);
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,str,NULL);
  return TCL_OK;
}


// ---- meanE : computes mean E
//

int ECData::meanE(const char* startMark, const char* endMark)
{
  ECRecord* rec=getECRecord(startMark);
  if (rec==NULL) return TCL_ERROR;

  if (strcmp(endMark,"")==0) endMark=startMark;

  int last = 0;
  int n = 0;
  double mean = 0;

  while (rec!=NULL)
    {
      if ((last==1)&&(strcmp(endMark,Tcl_DStringValue(rec->getMark()))!=0))
	break;

      rec->cpyECRecToTcl(interp,TclCmd);
      Tcl_Eval(interp,Tcl_DStringValue(cut));
      if (strcmp(interp->result,"1")==0)
	{
	  mean = mean+rec->getE();
	  n++;
	}

      if (strcmp(endMark,Tcl_DStringValue(rec->getMark()))==0) last=1;
      rec=rec->getNextECRecord();
    }

  mean = mean/n;
  char str[TCL_DOUBLE_SPACE];
  Tcl_PrintDouble(interp,mean,str);
  Tcl_ResetResult(interp);
  Tcl_AppendResult(interp,str,NULL);
  return TCL_OK;
}


// ---- cvQ : computes charge for a CV experiment
//            Numerical integration of abs(I)dE using trapezoidal rule

int ECData::cvQ(double scanRate, const char* startMark, const char* endMark)
{
  ECRecord* rec=getECRecord(startMark);
  if (rec==NULL) return TCL_ERROR;

  if (strcmp(endMark,"")==0) endMark=startMark;

  int last = 0;
  double Q = 0;
  double Qp = 0;
  double Qn = 0;

  while (rec->getNextECRecord()!=NULL)
    {
      if ((last==1)&&(strcmp(endMark,Tcl_DStringValue(rec->getMark()))!=0))
	break;

      rec->cpyECRecToTcl(interp,TclCmd);
      Tcl_Eval(interp,Tcl_DStringValue(cut));
      if (strcmp(interp->result,"1")==0)
	{
	  double E1=rec->getE();
	  double E2=rec->getNextECRecord()->getE();
	  double I1=rec->getI();
	  double I2=rec->getNextECRecord()->getI();

	  Q=Q+(E2-E1)*(abs(I1)+abs(I2))/2.0;
	  Qp=Qp+(E2-E1)*(max(0.0,I1)+max(0.0,I2))/2.0;
	  Qn=Qn-(E2-E1)*(min(I1,0.0)+min(I2,0.0))/2.0;
	}

      if (strcmp(endMark,Tcl_DStringValue(rec->getMark()))==0) last=1;
      rec=rec->getNextECRecord();
    }

  char str[TCL_DOUBLE_SPACE];
  Tcl_PrintDouble(interp,Q/scanRate,str);
  Tcl_SetVar2(interp,"cvQ","tot",str,TCL_GLOBAL_ONLY);
  Tcl_PrintDouble(interp,Qp/scanRate,str);
  Tcl_SetVar2(interp,"cvQ","pos",str,TCL_GLOBAL_ONLY);
  Tcl_PrintDouble(interp,Qn/scanRate,str);
  Tcl_SetVar2(interp,"cvQ","neg",str,TCL_GLOBAL_ONLY);
  Tcl_ResetResult(interp);
  Tcl_Eval(interp,"array get cvQ");
  return TCL_OK;
}
