// This may look like C code, but it is really -*- C++ -*-
// CVS ID : $Id: ECData.h,v 1.9 2011/03/08 01:33:28 wuthrich Exp $
// ---------------------------------------------------------------- 
//                                                                  
// Header file for  class ECData
//
// (c) Rolf Wuthrich
//     2010 Concordia University
//
// author:  Rolf Wuthrich
// email:   wuthrich@encs.concordia.ca
// version: v1 ($Revision: 1.9 $)
//
// This software is copyright under the BSD license
//
// ---------------------------------------------------------------
// history:
// ---------------------------------------------------------------
// Date       Modification                     Author
// ---------------------------------------------------------------
// 12.07.2010 Creation                         R. Wuthrich
// ---------------------------------------------------------------

#ifndef ECData_SEEN_
#define ECData_SEEN_

#include <tcl.h>
#include <fstream>
#include "ECObj.h"

using namespace std;


// -----------------------------------------------------------------------
//
// class ECRecord : a class for holding electrochemical data
//

class ECRecord
{

protected:

  double t;              // time [s]
  double E;              // potential [V]
  double I;              // current [A]
  double Q;              // charge (since started or reseted) [C]
  double Qpos;           // positive charge (since started or reseted) [C]
  double Qneg;           // negative charge (since started or reseted) [C]
  Tcl_DString* mark;     // mark used to identify specific time intervals

  Tcl_DString* prnStr;   // used for print method

  ECRecord*      nextECRecord;

public:

  ECRecord(double T, double potential, double current, 
           double charge, double PosCharge, double NegCharge,
	   const char* Mark);
  ECRecord(ECRecord* ecrec);
  ~ECRecord();

  void         setNextECRecord(ECRecord* next) {nextECRecord=next;}
  ECRecord*    getNextECRecord() {return nextECRecord;}

  double       getT() {return t;}
  double       getE() {return E;}
  double       getI() {return I;}
  double       getQ() {return Q;}
  double       getQpos() {return Qpos;}
  double       getQneg() {return Qneg;}
  void         setMark(Tcl_DString* Mark) {mark=Mark;}
  Tcl_DString* getMark() {return mark;}

  Tcl_DString* print(const char* separator = "\t");

  void cpyECRecToTcl(Tcl_Interp* interp, Tcl_DString* NameSpace);
};


// ------------------------------------------------------------------------
//
// class ECData : abstract class able to hold and process
//                EC data (a list of ECRecords)

class ECData : public virtual ECObj
{
private:

  Tcl_DString*   cut;             // used for cut method
  ECRecord*      currentECRecord;
  Tcl_Namespace* nsPtr;           // pointer to ECData Namespace

  int            setCut(int objc,Tcl_Obj *const objv[]);
  int            setXY(int objc,Tcl_Obj *const objv[]);
  void           parseMarks(int &i,int objc,Tcl_Obj* const objv[],
			    char* &begin, char* &end);

protected:

  Tcl_DString*   xCoord;          // used for x coordinate
  Tcl_DString*   yCoord;          // used for y coordinate
  Tcl_DString*   plotStyle;       // used for configurating plot

  ECRecord* firstECRecord;
  ECRecord* lastECRecord;

  int       n;      // number of records

  double    dt_out; // dt for output operations (plotting, saving, etc...) [s]

  char*     markStr;

public:

  ECData(const char *Name, Tcl_Interp *Interp);
  ~ECData();

  virtual int cmd(int &i,int objc,Tcl_Obj *const objv[]);
  virtual Tcl_DString* info(const char* frontStr = "");

  ECRecord*   getFirstECRecord() {return firstECRecord;};
  ECRecord*   getLastECRecord() {return lastECRecord;};
  ECRecord*   getECRecord(const char* Mark);
  void        addECRecord(double T, double potential, double current, 
			  double charge, double PosCharge, double NegCharge);
  void        removeRecords();
  void        recordsStat();
  virtual int getNbrECRecords();
  int         loop(const char* loopScript,
		   const char* startMark, const char* endMark);
  int         saveRecords(const char* fileName, ios_base::openmode mode,
			  const char* startMark, const char* endMark);
  int         scan(const char* startMark, const char* endMark);

  void        setMarkStr(const char* Mark);
  double      dtOut();

  int         findMax(const char* startMark, const char* endMark);
  int         findMin(const char* startMark, const char* endMark);
  int         mean(const char* startMark, const char* endMark);
  int         meanI(const char* startMark, const char* endMark);
  int         meanE(const char* startMark, const char* endMark);
  int         cvQ(double scanRate, const char* startMark, const char* endMark);
};


#endif /* ECData_SEEN_ */
