// CVS ID : $Id: ECObj.cc,v 1.6 2011/03/08 01:33:28 wuthrich Exp $
// ---------------------------------------------------------------
//
// Definition of the class ECObj
//
// (c) Rolf Wuthrich
//     2010 Concordia University
//
// author:  Rolf Wuthrich
// email:   wuthrich@encs.concordia.ca
// version: v1 ($Revision: 1.6 $)
//
// This software is copyright under the BSD license
//  
// ---------------------------------------------------------------
// history:
// ---------------------------------------------------------------
// Date       Modification                     Author
// ---------------------------------------------------------------
// 09.07.2010 Creation                         R. Wuthrich
// ---------------------------------------------------------------

#include <cstdlib>
#include <cstring>
#include <tcl.h>
#include "ECObj.h"

using namespace std;

int ECObj::IDcounter = 0;


// -------------------------------------------------------------------------
//
// Tcl interface


// Command to manipulate an existing ECObj object
//

int ECTk_ECObjCmd(ClientData clientData, Tcl_Interp *interp, 
		  int objc,Tcl_Obj *const objv[])
{
  ECObj* ecobj=(ECObj *)clientData;

  if (objc==1)
    {
      Tcl_AppendResult(interp,ecobj->getID(),NULL);
      return TCL_OK;
    }

  int i=1;
  while (i<objc)
    if (ecobj->cmd(i,objc,objv)==-1) return TCL_ERROR;

  return TCL_OK;
}


// -----------------------------------------------------------------------
//
// class ECObj : abstract class from which all EC objects derive
//


// ---- CONSTRUCTOR
//

ECObj::ECObj(const char* objName, const char* objCmd, Tcl_Interp* Interp) :
  name(objName),
  interp(Interp),
  cmdOpts(NULL)
{
  IDcounter++;
  sprintf(objID,"%d",IDcounter);
  TclCmd  = new Tcl_DString;
  infoStr = new Tcl_DString;
  Tcl_DStringInit(TclCmd);
  Tcl_DStringAppend(TclCmd,objCmd,-1);
  Tcl_DStringAppend(TclCmd,objID,-1);
  Tcl_DStringInit(infoStr);

  addOption("-help","displays this message");
  addOption("-name","returns name of process");
  addOption("-info","returns info");
  addOption("-nRecords","returns number of records");
}


// ---- DESTRUCTOR
//

ECObj::~ECObj()
{
  Tcl_DeleteCommand(interp,Tcl_DStringValue(TclCmd));
  Tcl_DStringFree(TclCmd);
  Tcl_DStringFree(infoStr);
  delete TclCmd;
  delete infoStr;
  Tcl_UnsetVar2(interp,"::ectk::procList",objID,0);
  Tcl_UnsetVar2(interp,"::ectk::recList",objID,0);
  int i=0;
  while (cmdOpts[i]!=NULL)
    {
      Tcl_DStringFree(assVars[i].help);
      delete assVars[i].help;
      delete cmdOpts[i];
      i++;
    }
}


// ---- ADDOPTION : adds an option to the option table
//

int ECObj::addOpt(const char* opt)
{
  int i=0;

  if (cmdOpts!=NULL)
    while (cmdOpts[i]!=NULL)
      i++;

  cmdOpts = (char **)realloc(cmdOpts,(i+2) * sizeof *cmdOpts);
  cmdOpts[i] = (char *)malloc(strlen(opt) + 1);
  strcpy(cmdOpts[i],opt);
  cmdOpts[i+1]=NULL;
  return i;
}

int ECObj::addOption(const char* opt, const char* help)
{
  int i=addOpt(opt);
  AssocVar av;
  av.varPtr=NULL;
  av.varType=EC_NO_VAR;
  av.help= new Tcl_DString;
  Tcl_DStringInit(av.help);
  Tcl_DStringAppend(av.help,help,-1);
  assVars.push_back(av);
  return i;
}

int ECObj::addOption(const char* opt, double* dblPtr, const char* help)
{
  int i=addOpt(opt);
  AssocVar av;
  av.varPtr=dblPtr;
  av.varType=EC_DOUBLE;
  av.help= new Tcl_DString;
  Tcl_DStringInit(av.help);
  Tcl_DStringAppend(av.help,help,-1);
  assVars.push_back(av);
  return i;
}

int ECObj::addOption(const char* opt, int* intPtr, const char* help)
{
  int i=addOpt(opt);
  AssocVar av;
  av.varPtr=intPtr;
  av.varType=EC_INT;
  av.help= new Tcl_DString;
  Tcl_DStringInit(av.help);
  Tcl_DStringAppend(av.help,help,-1);
  assVars.push_back(av);
  return i;
}

int ECObj::addOption(const char* opt, Tcl_DString* DStrPtr, const char* help)
{
  int i=addOpt(opt);
  AssocVar av;
  av.varPtr=DStrPtr;
  av.varType=EC_DSTRING;
  av.help= new Tcl_DString;
  Tcl_DStringInit(av.help);
  Tcl_DStringAppend(av.help,help,-1);
  assVars.push_back(av);
  return i;
}


// ---- GETOPTIONINDEX : returns the entry number from the options table
//

int ECObj::getOptionIndex(const char* opt)
{
  int i;
  Tcl_GetIndexFromObj(NULL,Tcl_NewStringObj(opt,-1),optsTbl(),"",0,&i);
  return i;
}


// ---- OPTSTBL : returns the adress of the options table
//

const char** ECObj::optsTbl()
{
  return (const char **)cmdOpts;
}


// ---- CMD : process a potential command option (# i of objv[])
//            searches the options table 
//            if a match is found processes the command option and
//            returns the index from the options table
//            if no match returns -1

int ECObj::cmd(int &i, int objc,Tcl_Obj *const objv[])
{
  char st[80];
  int index=-1;
  int j=0;
  Tcl_GetIndexFromObj(interp,objv[i],optsTbl(),"option",0,&index);
  if (index==-1) return -1;

  // handles the options -help, -info, -name and -nRecords
  switch (index)
    {

    case 0: // help
      Tcl_AppendResult(interp,"\n",Tcl_GetString(objv[0])," - ",name,NULL);
      Tcl_AppendResult(interp,"\n\nValid options are:\n\n",NULL);
      while (cmdOpts[j]!=NULL)
	{
	  sprintf(st,"%-20.20s %-58.58s\n",cmdOpts[j],
		  Tcl_DStringValue(assVars[j].help));
	  Tcl_AppendResult(interp,st,NULL);
	  j++;
	}
      Tcl_AppendResult(interp,"\nFor more details see man page.",NULL);
      i++;
      break;

    case 1: // name
      Tcl_AppendResult(interp,name,NULL);
      i++;
      break;

    case 2: // info
      Tcl_DStringResult(interp,info());
      i++;
      break;

    case 3: // nRecords
      char str[TCL_DOUBLE_SPACE];
      sprintf(str,"%i",getNbrECRecords());
      Tcl_AppendResult(interp,str,NULL);
      i++;
      break;

    }

  // handles the options with associated variables
  char str[TCL_DOUBLE_SPACE];
  double* pd;
  int* pi;
  Tcl_DString* pstr;

  switch (assVars[index].varType)
    {

    case EC_DOUBLE:
      pd = (double*)assVars[index].varPtr;
      if (i+1<=objc-1)
	{
	  if (Tcl_GetDoubleFromObj(interp,objv[i+1],pd)!=TCL_OK) return -1;
	  i++;
	}
      else
	{
	  Tcl_PrintDouble(interp,*pd,str);
	  Tcl_AppendResult(interp,str,NULL);
	}
      i++;
      break;

    case EC_INT:
      pi = (int*)assVars[index].varPtr;
      if (i+1<=objc-1)
	{
	  if (Tcl_GetIntFromObj(interp,objv[i+1],pi)!=TCL_OK) return -1;
	  i++;
	}
      else
	{
	  sprintf(str,"%d",*pi);
	  Tcl_AppendResult(interp,str,NULL);
	}
      i++;
      break;

    case EC_DSTRING:
      pstr = (Tcl_DString *)assVars[index].varPtr;
      if ((i+1<=objc-1)&&(Tcl_StringMatch(Tcl_GetString(objv[i+1]),"-*")==0))
	{
	  Tcl_DStringFree(pstr);
	  Tcl_DStringAppend(pstr,Tcl_GetString(objv[i+1]),-1);
	  i=i+2;
	}
      else
	{
	  Tcl_AppendResult(interp,Tcl_DStringValue(pstr),NULL);
	  i++;
	}
      break;

    }

  return index;
}


// ---- GETTCLCMD : returns the associated TclCmd
//

const char* ECObj::getTclCmd()
{
  return Tcl_DStringValue(TclCmd);
}


// ---- INFO : returns a Tcl_DString containing relvant info about content
//             Has to be completed by child

Tcl_DString* ECObj::info(const char* frontStr)
{
  Tcl_DStringFree(infoStr);
  Tcl_DStringAppend(infoStr,frontStr,-1);
  Tcl_DStringAppend(infoStr,name,-1);
  Tcl_DStringAppend(infoStr,":",-1);
  return infoStr;
}


// ---- ADDINFO : adds a line to the indoStr (for a string variable)
//

void ECObj::addInfo(const char* frontStr, const char* text, const char* var)
{
  Tcl_DStringAppend(infoStr,"\n",-1);
  Tcl_DStringAppend(infoStr,frontStr,-1);
  Tcl_DStringAppend(infoStr,text,-1);
  Tcl_DStringAppend(infoStr,var,-1);
}


// ---- ADDINFO : adds a line to the indoStr (for a int variable)
//

void ECObj::addInfo(const char* frontStr, const char* text, int var)
{
  char str[TCL_DOUBLE_SPACE];
  Tcl_DStringAppend(infoStr,"\n",-1);
  Tcl_DStringAppend(infoStr,frontStr,-1);
  Tcl_DStringAppend(infoStr,text,-1);
  sprintf(str,"%i",var);
  Tcl_DStringAppend(infoStr,str,-1);
}


// ---- ADDINFO : adds a line to the indoStr (for a double variable)
//

void ECObj::addInfo(const char* frontStr, const char* text, double var)
{
  char str[TCL_DOUBLE_SPACE];
  Tcl_DStringAppend(infoStr,"\n",-1);
  Tcl_DStringAppend(infoStr,frontStr,-1);
  Tcl_DStringAppend(infoStr,text,-1);
  Tcl_PrintDouble(interp,var,str);
  Tcl_DStringAppend(infoStr,str,-1);
}


// ---- GETID : returns the variable objID;
//

const char* ECObj::getID()
{
  return objID;
}


// ---- GETECOBJNAME : returns the variable name;
//

const char* ECObj::getECObjName()
{
  return name;
}


// ---- GETNBRECRECORDS : returns -1
//                        can be used to distinguish 
//                        an ECProcess from an ECProcedure

int ECObj::getNbrECRecords()
{
  return -1;
}
