// CVS ID : $Id: ECProcess.cc,v 1.9 2010/12/05 20:14:48 wuthrich Exp $
// ---------------------------------------------------------------
//
// Definition of the class ECProcess
//
// (c) Rolf Wuthrich
//     2010 Concordia University
//
// author:  Rolf Wuthrich
// email:   wuthrich@encs.concordia.ca
// version: v1 ($Revision: 1.9 $)
//
// This software is copyright under the BSD license
//  
// ---------------------------------------------------------------
// history:
// ---------------------------------------------------------------
// Date       Modification                     Author
// ---------------------------------------------------------------
// 31.03.2010 Creation                         R. Wuthrich
// 20.06.2010 Added getNbrECRecords()          R. Wuthrich
// ---------------------------------------------------------------

#include <cstring>
#include <sys/time.h>
#include <tcl.h>
#include "ECProcess.h"
#include "ECApp.h"

using namespace std;


// -------------------------------------------------------------------------
//
// Tcl interface


// Command to create a new ECProcess object
//

int ECTk_CreateECProcessCmd(Tcl_Interp *interp, ClientData clientData,
			    ECProcess* proc, int objc, Tcl_Obj *const objv[])
{
  ECApp* app = (ECApp *)clientData;

  Tcl_CreateObjCommand(interp,proc->getTclCmd(),
		       ECTk_ECObjCmd,static_cast<ECObj*>(proc),
		       (Tcl_CmdDeleteProc *) NULL);

  // checks if -help option was used
  int i;
  Tcl_GetIndexFromObj(NULL,objv[1],proc->optsTbl(),"",0,&i);
  if (i==proc->getOptionIndex("-help"))
    {
      ECTk_ECObjCmd((ClientData)proc,interp,2,objv);
      delete proc;
      return TCL_OK;
    }

  // invokes the new command in order to manipulate the ECTk object
  // in case any error occurs, the object is deleted as 
  // well as its associated Tcl command

  if (ECTk_ECObjCmd((ClientData)proc,interp,objc,objv)==TCL_OK)
    {
      app->addECProcess(proc);
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,proc->getID(),NULL);
      return TCL_OK;
    }
  else
    {
      delete proc;
      return TCL_ERROR;
    }
}


// -----------------------------------------------------------------------
//
// class ECProcess : abstract class from which all EC processes derive
//


// ---- CONSTRUCTOR
//

ECProcess::ECProcess(const char* procName, const char* procOwner,
		     const char* procCmd, Tcl_Interp* Interp) : 
  ECObj(procName,procCmd,Interp),
  nextECProc(NULL),
  owner(procOwner),
  status(Waiting),
  verbose(1)
{
  activationTime.tv_sec=-1;
  activationTime.tv_usec=-1;
  preProcessScript  = new Tcl_DString;
  postProcessScript = new Tcl_DString;
  Tcl_DStringInit(preProcessScript);
  Tcl_DStringInit(postProcessScript);

  addOption("-quiet","turns off messaging to ECTk console");
  addOption("-verbose","turns on messaging to ECTk console");
  addOption("-preProcess",preProcessScript,"returns/sets pre-process script");
  addOption("-postProcess",postProcessScript,"returns/sets post-process script");
}


// ---- DESTRUCTOR
//

ECProcess::~ECProcess()
{
  Tcl_DStringFree(preProcessScript);
  Tcl_DStringFree(postProcessScript);
  delete preProcessScript;
  delete postProcessScript;
}


// ---- CMD : process a potential command option (# i of objv[])
//            searches the options table 
//            if a match is found processes the command option and
//            returns the index from the options table
//            if no match returns -1

int ECProcess::cmd(int &i, int objc,Tcl_Obj *const objv[])
{
  // first executes the command options defined in ECObj
  int index=ECObj::cmd(i,objc,objv);

  if (index==getOptionIndex("-quiet"))
    {
      verbose=0;
      i++;
    }

  if (index==getOptionIndex("-verbose"))
    {
      verbose=1;
      i++;
    }

  return index;
}


// ---- HANDLEEVENT : defines how the ECProcess reacts to external events
//                    Has to be completed by child

void ECProcess::handleEvent(ECEvent* ev)
{
  if ((ev->cmd()==CmdStop)) status=Terminated;
}


// ---- INFO : returns a Tcl_DString containing relevant info about content
//             Has to be completed by child

Tcl_DString* ECProcess::info(const char* frontStr)
{
  ECObj::info(frontStr);
}


// ---- GETNBRECRECORDS : returns -1
//                        can be used to distinguish 
//                        an ECProcess from an ECProcedure

int ECProcess::getNbrECRecords()
{
  return -1;
}


// ---- TERMINATE : called when the process is terminated:
//                  - sets status to terminate 
//                  - executes the postProcess script

void ECProcess::terminate()
{
  status=Terminated;
  char str[80];
  if (verbose==1)
    {
      sprintf(str,"cons \"terminated %s ($::ectk::procList(%s)).\"",
	      name,objID);
      Tcl_Eval(interp,str);
    }
  Tcl_SetVar(interp,"ID",objID,0);
  Tcl_Eval(interp,Tcl_DStringValue(postProcessScript));
}


// ---- ACTIVATE : called when the process is activated:
//                 - sets status to Active 
//                 - executes the preProcess script
//                 - sets the activation time

void ECProcess::activate(ECEvent* ev)
{
  status=Active;
  ev->setT(0);
  ev->resetQ();
  char str[80];
  if (verbose==1)
    {
      sprintf(str,"cons \"\n==> starting %s ($::ectk::procList(%s)).\"",
	      name,objID);
      Tcl_Eval(interp,str);
    }
  Tcl_SetVar(interp,"ID",objID,0);
  Tcl_Eval(interp,Tcl_DStringValue(preProcessScript));
  gettimeofday(&activationTime,NULL);
}


// ---- TIMESINCEACTIVATED : returns the time elapsed since activated
//

double ECProcess::timeSinceActivated()
{
  struct timeval tv;
  gettimeofday(&tv,NULL);
  double t=tv.tv_sec-activationTime.tv_sec+
    (tv.tv_usec-activationTime.tv_usec)*1e-6;
  return t;
}


// ---- SETPREPROCESSSCRIPT : sets the variable preProcessScript
//

void ECProcess::setPreProcessScript(Tcl_DString* PreProcessScript)
{
  Tcl_DStringFree(preProcessScript);
  Tcl_DStringAppend(preProcessScript,Tcl_DStringValue(PreProcessScript),-1);
}


// ---- SETPOSTPROCESSSCRIPT : sets the variable postProcessScript
//

void ECProcess::setPostProcessScript(Tcl_DString* PostProcessScript)
{
  Tcl_DStringFree(postProcessScript);
  Tcl_DStringAppend(postProcessScript,Tcl_DStringValue(PostProcessScript),-1);
}


// ---- GETPREPROCESSSCRIPT : sets the variable preProcessScript
//

Tcl_DString* ECProcess::getPreProcessScript()
{
  return preProcessScript;
}


// ---- GETPOSTPROCESSSCRIPT : sets the variable postProcessScript
//

Tcl_DString* ECProcess::getPostProcessScript()
{
  return postProcessScript;
}
 

// ---- GETPROCESSOWNER : returns the variable owner;
//

const char* ECProcess::getProcessOwner()
{
  return owner;
}


// ---- GETSTATUS : returns the variable status
//

int ECProcess::getStatus()
{
  return status;
}


// ---- SETSTATUS : sets the process status
//

void ECProcess::setStatus(int ECProcessStatus)
{
  status=ECProcessStatus;
}


// ---- GETSTATUSSTR : returns the process status
//

const char* ECProcess::getStatusStr()
{
  return StatusStr[status];
}


// ---- SETNEXTECOBJ : assigns the variable nextECObj
//

void ECProcess::setNextECProcess(ECProcess *NextECProc)
{
  nextECProc=NextECProc;
}


// ---- GETNEXTECPROCESS : returns the variable nextECProc
//

ECProcess* ECProcess::getNextECProcess()
{
  return nextECProc;
}
