// CVS ID : $Id: ECRec.cc,v 1.5 2010/12/06 00:53:58 wuthrich Exp $
// ---------------------------------------------------------------
//
// Definition of the class ECRecord
//
// (c) Rolf Wuthrich
//     2010 Concordia University
//
// author:  Rolf Wuthrich
// email:   wuthrich@encs.concordia.ca
// version: v1 ($Revision: 1.5 $)
//
// This software is copyright under the BSD license
//  
// ---------------------------------------------------------------
// history:
// ---------------------------------------------------------------
// Date       Modification                     Author
// ---------------------------------------------------------------
// 27.06.2010 Creation                         R. Wuthrich
// ---------------------------------------------------------------

#include <tcl.h>
#include <cstring>
#include "ECRec.h"
#include "ECApp.h"

using namespace std;


// -------------------------------------------------------------------------
//
// Tcl interface


// Command to create a new ecrec object
// 

int ECTk_ECRecCmd(ClientData clientData, Tcl_Interp *interp, 
		  int objc,Tcl_Obj *const objv[])
{
  ECApp* app=(ECApp *)clientData;

  // creates a new Tcl command that can be used later 
  // to manipulate the ECRec object

  if (objc==1)
    {
      Tcl_WrongNumArgs(interp,1,objv,"subcommand ?argument ...?");
      return TCL_ERROR;
    }

  char cmd[80];
  int n=0;
  Tcl_CmdInfo cmdInfo;
  if (objc>2)
    if (Tcl_GetCommandInfo(interp,Tcl_GetString(objv[2]),&cmdInfo)==1)
      {
	Tcl_AppendResult(interp,"error: \"",Tcl_GetString(objv[2]),
			 "\" already exists",NULL);
	return TCL_ERROR;
      }
    else
      strcpy(cmd,Tcl_GetString(objv[2]));

  int index;
  static CONST char *opts[] = {"-copy","-load",NULL}; 

  if (Tcl_GetIndexFromObj(interp,objv[1],opts,"subcommand",0,&index)!=TCL_OK)
    return TCL_ERROR;

  ECProcedure* p;
  ECRec* rec = new ECRec(cmd,interp);

  switch (index)
    {
    case 0: // -copy
      if (objc<4)
	{
	  Tcl_WrongNumArgs(interp,2,objv,"ECRec PID ?startMark? ?endMark?");
	  goto ERROR;
	}
      p=app->findECProcedure(Tcl_GetString(objv[3]));
      if (p==NULL)
	{
	  Tcl_AppendResult(interp,"invalid PID or not an ECProcedure",NULL);
	  goto ERROR;
	}

      switch (objc)
	{
	case 4:
	  n=rec->copyRecords(p);
	  break;
	case 5:
	  n=rec->copyRecords(p,Tcl_GetString(objv[4]));
	  break;
	case 6:
	  n=rec->copyRecords(p,Tcl_GetString(objv[4]),Tcl_GetString(objv[5]));
	  break;
	default:
	  Tcl_WrongNumArgs(interp,2,objv,"ECRec PID ?startMark? ?endMark?");
	  goto ERROR;
	  break;
	}
      break;

    case 1: // -load
      switch (objc)
	{
	case 4:
	  n=rec->loadRecords(Tcl_GetString(objv[3]));
	  break;
	case 5:
	  n=rec->loadRecords(Tcl_GetString(objv[3]),Tcl_GetString(objv[4]));
	  break;
	case 6:
	  n=rec->loadRecords(Tcl_GetString(objv[3]),Tcl_GetString(objv[4]),Tcl_GetString(objv[5]));
	  break;
	default:
	  Tcl_WrongNumArgs(interp,2,objv,"ECRec fileName ?startMark? ?endMark?");
	  goto ERROR;
	  break;
	}

      if (n==-1)
	{
	  Tcl_AppendResult(interp,"couldn't open \"",Tcl_GetString(objv[3]),
			          "\": no such file",NULL);
	  goto ERROR;
	}
      break;
    }

  Tcl_CreateObjCommand(interp,cmd,ECTk_ECObjCmd, 
		       static_cast<ECObj*>(rec), (Tcl_CmdDeleteProc *) NULL);

  app->addECRecord(rec,cmd);
  sprintf(cmd,"%d",n);
  Tcl_AppendResult(interp,cmd,NULL);
  return TCL_OK;

 ERROR:
  delete rec;
  return TCL_ERROR;
}


// -----------------------------------------------------------------------
//
// class ECRec : class able to hold and manipulate EC data
//


// ---- CMD : process a potential command option (# i of objv[])
//            searches the options table 
//            if a match is found processes the command option and
//            returns the index from the options table
//            if no match returns -1

int ECRec::cmd(int &i, int objc,Tcl_Obj *const objv[])
{
  // executes the command options defined in ECProcedure
  return ECData::cmd(i,objc,objv);
}


// ---- INFO : returns a Tcl_DString containing relevant info about content
//

Tcl_DString* ECRec::info(const char* frontStr)
{
  ECData::info(frontStr);

  char str[80];
  char mar[80];
  char dbl[TCL_DOUBLE_SPACE];

  Tcl_DStringAppend(infoStr,frontStr,-1);
  Tcl_DStringAppend(infoStr,"\n--------------------------------",-1);
  addInfo(frontStr,"Entries = ",n);
  Tcl_DStringAppend(infoStr,"\n",-1);
  Tcl_DStringAppend(infoStr,frontStr,-1);
  Tcl_DStringAppend(infoStr,"--------------------------------\n",-1);
  Tcl_DStringAppend(infoStr,frontStr,-1);
  Tcl_DStringAppend(infoStr,"Variable\tLower\tUpper\n",-1);
  Tcl_DStringAppend(infoStr,frontStr,-1);
  Tcl_DStringAppend(infoStr,"--------------------------------\n",-1);

  ECRecord* rec=firstECRecord;
  if (rec==NULL) return infoStr;

  double minT = rec->getT();
  double maxT = rec->getT();
  double minE = rec->getE();
  double maxE = rec->getE();
  double minI = rec->getT();
  double maxI = rec->getT();
  double minQ = rec->getE();
  double maxQ = rec->getE();

  sprintf(mar,"%s",Tcl_DStringValue(rec->getMark()));
  Tcl_PrintDouble(interp,rec->getT(),dbl);

  while (rec->getNextECRecord()!=NULL)
    {
      if (rec->getT()>maxT) maxT=rec->getT();
      if (rec->getT()<minT) minT=rec->getT();
      if (rec->getE()>maxE) maxE=rec->getE();
      if (rec->getE()<minE) minE=rec->getE();
      if (rec->getI()>maxI) maxI=rec->getI();
      if (rec->getI()<minI) minI=rec->getI();
      if (rec->getQ()>maxQ) maxQ=rec->getQ();
      if (rec->getQ()<minQ) minQ=rec->getQ();
      rec=rec->getNextECRecord();
    }
  sprintf(str,"%st [s]\t    %9.3f %9.3f\n",frontStr,minT,maxT);
  Tcl_DStringAppend(infoStr,str,-1);
  sprintf(str,"%sE [V]\t    %9.3f %9.3f\n",frontStr,minE,maxE);
  Tcl_DStringAppend(infoStr,str,-1);
  sprintf(str,"%sI [A]\t    %9.3g %9.3g\n",frontStr,minI,maxI);
  Tcl_DStringAppend(infoStr,str,-1);
  sprintf(str,"%sQ [C]\t    %9.3g %9.3g\n",frontStr,minQ,maxQ);
  Tcl_DStringAppend(infoStr,str,-1);
  sprintf(str,"%sMarks\t    %9s %9s\n",frontStr,mar,Tcl_DStringValue(rec->getMark()));
  Tcl_DStringAppend(infoStr,str,-1);
  Tcl_DStringAppend(infoStr,"--------------------------------\n",-1);

  return infoStr;
}


// ---- LOADRECORDS : loads data from a file to an ECRecord
//                    if startMark = "" starts from the first record
//

int ECRec::loadRecords(const char* fileName,
		       const char* startMark, const char* endMark)
{
  ifstream file;
  file.open(fileName);
  if (!file) 
    {
      file.close();
      return -1;
    }

  int rec = 0;
  int last = 0;
  if (strcmp(endMark,"")==0)
    endMark=startMark;

  while (!file.eof())
    {
      double t,E,I,Q,Qpos,Qneg;
      char mark[80];
      file >> t >> E >> I >> Q >> Qpos >> Qneg >> mark;
      if (file.fail())
	{
	  file.clear();
	  file.get();
	}
      else
	{
	  if ((strcmp(mark,startMark)==0)||(strcmp(startMark,"")==0)) rec=1;
	  if (strcmp(mark,endMark)==0) last=1;
 	  if ((strcmp(mark,endMark)!=0)&&(last==1)) rec=0;
	  setMarkStr(mark);
	  if (rec==1) addECRecord(t,E,I,Q,Qpos,Qneg);
	}
    }
  file.close();
  return n;
}


// ---- COPYRECORDS : fills EC data by copying the data from an existing
//                    ECTk object
//
//                    if startMark = "" starts from the first record
//

int ECRec::copyRecords(ECProcedure* ecproc,
		       const char* startMark, const char* endMark)
{
  int rec  = 0;
  int last = 0;
  char mark[8];
  if (strcmp(endMark,"")==0)
    endMark=startMark;

  ECRecord* ECRec = ecproc->getFirstECRecord();
  ECRecord* NewECRec;
  firstECRecord = NULL;

  while (ECRec!=NULL)
    {
      strcpy(mark,Tcl_DStringValue(ECRec->getMark()));
      if (((strcmp(mark,startMark)==0)||(strcmp(startMark,"")==0))&&
	  (firstECRecord==NULL))
	{
	  rec=1;
	  NewECRec = new ECRecord(ECRec);
	  n++;
	  firstECRecord=NewECRec;
	  ECRec=ECRec->getNextECRecord();
	  continue;
	}
      if (strcmp(mark,endMark)==0) last=1;
      if ((strcmp(mark,endMark)!=0)&&(last==1)) rec=0;

      if (rec==1)
	{
	  NewECRec->setNextECRecord(new ECRecord(ECRec));
	  NewECRec=NewECRec->getNextECRecord();
	  lastECRecord=NewECRec;
	  n++;
	}
      ECRec=ECRec->getNextECRecord();
    }
  return n;
}
